/*
 * Copyright (C) 2002  Pekka Enberg <Pekka.Enberg@cs.helsinki.fi>
 *
 * Distributed under the terms of the GNU General Public License
 * version 2 or later.
 */
#include <errno.h>
#include <stdarg.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* Print error message.  */
void error(const char * format, ...)
{
    va_list args;

    fprintf(stderr, "Error: ");

    va_start(args, format);
    vfprintf(stderr, format, args);
    va_end(args);

    fputc('\n', stderr);
}

#define INDENT_CHAR      ' '
#define NUM_INDENT_CHARS 4

/* Print indent characters.  */
void print_indent(FILE * output, int indent_level)
{
    int i;

    for (i = 0; i < (NUM_INDENT_CHARS * indent_level); i++) {
	fputc(INDENT_CHAR, output);
    }
}

/* Check if for XML begin tag.  */
static inline bool is_xml_begin_tag(int character, int lookahead)
{
    return ('<' == character) && ('/' != lookahead);
}

/* Check for XML end tag.  */
static inline bool is_xml_end_tag(int character, int lookahead)
{
    return ((('<' == character) && ('/' == lookahead)) ||
	    (('/' == character) && ('>' == lookahead)));
}

/* Indent XML buffer.  */
void indent_buffer(const char * buffer, size_t count, int indent_level,
		   FILE * output)
{
    size_t pos;
    size_t last_pos;
    bool begin_tag_found;

    last_pos = 0;
    begin_tag_found = false;

    for (pos = 0; pos < count; pos++) {
	int c = buffer[pos];
	int lookahead = buffer[pos + 1];

	if (is_xml_begin_tag(c, lookahead)) {
	    /* XML begin tag changes indent level only after we have
               finished the current line.  */
	    begin_tag_found = true;
	} else if (is_xml_end_tag(c, lookahead)) {
	    if (begin_tag_found) {
		/* Begin and end tag on the same line balance each other.  */
		begin_tag_found = false;
	    } else {
		indent_level--;
		if (indent_level < 0) {
		    error("XML is not well-formed.");
		    indent_level = 0;
		}
	    }
	}

	if ('\n' == c) {
	    print_indent(output, indent_level);
	    if (fwrite(buffer + last_pos, sizeof(char), pos - last_pos + 1,
		       output) != pos - last_pos + 1) {
		error("Could not write to output stream: %s.",
		      strerror(errno));
	    }
	    last_pos = pos + 1;

	    if (begin_tag_found) {
		indent_level++;
		begin_tag_found = false;
	    }
	}
    }
}

#define MAX_BUFFER_LENGTH 1024

/* Indent XML stream.  */
void indent(FILE * input, FILE * output)
{
    int indent_level;
    char buffer[MAX_BUFFER_LENGTH];
    size_t count;

    indent_level = 0;

    for (;;) {
	count = fread((void *) buffer, sizeof(char), MAX_BUFFER_LENGTH, input);
	if (0 == count) {
	    if (ferror(input) != 0) {
		error("Error reading input stream: %s.", strerror(errno));
	    }
	    break;
	}

	indent_buffer(buffer, count, indent_level, output);
    }
}

int main(int argc, char * argv[])
{
    FILE * input;

    if (argc > 1) {
	input = fopen(argv[1], "r");
	if (NULL == input) {
	    error("Could not open '%s': %s.", argv[1], strerror(errno));
	    return EXIT_FAILURE;
	}
    } else {
	input = stdin;
    }

    indent(input, stdout);
    fclose(input);

    return EXIT_SUCCESS;
}
